/**********************************************************
 * Outbound management.
 **********************************************************/
/*
 * $Id: outbound.c,v 1.8 2005/02/09 12:44:23 mitry Exp $
 *
 * $Log: outbound.c,v $
 * Revision 1.8  2005/02/09 12:44:23  mitry
 * Changed free() to xfree()
 * Changed out_bso_rescan() code
 *
 * Revision 1.7  2005/02/08 19:57:28  mitry
 * Added outbound_addr_busy().
 * Wiped out old outbound_rescan code.
 * Cleaned code a bit.
 *
 *
 */

#include "headers.h"

#if defined(__linux__)
#  define OCONST const
#else
#  define OCONST
#endif


static int bso_defzone = 2;
static char tmp[MAX_STRING];
static char *aso_base, *bso_base, *sts_base, *p_domain;
static int pd_len = 0;

static void (*_each)(char *, ftnaddr_t *, int, int, int);
static int _rslow;


static char *ext_lo[] = {
    ".pnt", ".req",
    ".ilo", ".clo", ".dlo", ".flo", ".hlo",
    NULL
};

static char *ext_ut[] = {
    ".iut", ".cut", ".dut", ".out", ".hut",
    NULL
};


int outbound_init(char *asopath,char *bsopath,char *stspath,int def_zone)
{
    int rc = 0;

    if ( bsopath && *bsopath ) {
        char *p = strrchr( bsopath, '/' );

        if ( !(p + 1)) {
            *p = '\0';
            p = strrchr( bsopath, '/' );
        }

        if ( p ) {
            p_domain = xstrdup( p + 1 );
            pd_len = strlen( p_domain );
            *p = 0;
            bso_base = xstrdup( bsopath );
            bso_defzone = def_zone;
            rc |= BSO;
        } else
            bsopath = NULL;
    }

    if ( !bsopath ) {
        bso_base = NULL;
        p_domain = NULL;
    }

    if ( asopath && *asopath ) {
        int len = strlen( asopath ) - 1;
        if ( asopath[ len ] == '/' )
            asopath[ len ] = '\0';
        aso_base = xstrdup( asopath );
        rc |= ASO;
    } else
        aso_base = NULL;

    if ( stspath && *stspath ) {
        int len = strlen( stspath ) - 1;
        if ( stspath[ len ] == '/' )
            stspath[ len ] = '\0';
        sts_base = xstrdup( stspath );
    } else if ( bsopath )
        sts_base = xstrdup( bsopath );
    else if ( asopath )
        sts_base = xstrdup( asopath );

    return rc;
}


void outbound_done(void)
{
    xfree( aso_base );
    xfree( bso_base );
    xfree( sts_base );
    xfree( p_domain );
}


static char *get_outbound_name(int type, ftnaddr_t *fa)
{
    if ( type == ASO ) {
        snprintf( tmp, MAX_STRING, "%s/%d.%d.%d.%d.",
            aso_base, fa->z, fa->n, fa->f, fa->p);
    } else {
        char t[30];
        snprintf( tmp, MAX_STRING, "%s/%s", bso_base, p_domain );
        if ( fa->z != bso_defzone ) {
            snprintf( t, 30, ".%03x", fa->z );
            xstrcat( tmp, t, MAX_STRING - strlen( tmp ));
        }
        snprintf( t, 30, "/%04x%04x.", fa->n, fa->f );
        xstrcat( tmp, t, MAX_STRING - strlen( tmp ));
        if( fa->p ) {
            snprintf( t, 30, "pnt/%08x.", fa->p );
            xstrcat( tmp, t, MAX_STRING  - strlen( tmp ));
        }
    }
    return tmp;
}

static char *aso_pktn(int type, ftnaddr_t *fa, int fl)
{
    char *e = "eut";
    get_outbound_name( type, fa );
    switch(fl) {
        case F_NORM:
        case F_REQ: e = "out"; break;
        case F_DIR: e = "dut"; break;
        case F_CRSH:e = "cut"; break;
        case F_HOLD:e = "hut"; break;
        case F_IMM: e = "iut"; break;
    }
    xstrcat( tmp, e, MAX_STRING );
    return tmp;
}

static char *aso_flon(int type,ftnaddr_t *fa, int fl)
{
    char *e = "elo";
    get_outbound_name( type, fa );
    switch( fl ) {
        case F_NORM:
        case F_REQ: e = "flo"; break;
        case F_DIR: e = "dlo"; break;
        case F_CRSH:e = "clo"; break;
        case F_HOLD:e = "hlo"; break;
        case F_IMM: e = "ilo"; break;
    }
    xstrcat( tmp, e, MAX_STRING );
    return tmp;
}


static char *get_outbound_busy_name(int type, ftnaddr_t *fa, char b)
{
    char bn[] = "esy";
    
    *bn = b;
    get_outbound_name( type, fa );
    xstrcat( tmp, bn, MAX_STRING );
    return tmp;
}


static char *aso_reqn(int type,ftnaddr_t *fa)
{
    get_outbound_name( type, fa );
    xstrcat( tmp, "req", MAX_STRING );
    return tmp;
}


static char *aso_stsn(ftnaddr_t *fa)
{
    snprintf( tmp, MAX_STRING, "%s/%d.%d.%d.%d.qst",
        sts_base, fa->z, fa->n, fa->f, fa->p);
    return tmp;
}


static void out_scan_flo(char *flo_name, ftnaddr_t *addr)
{
    int len = strlen( flo_name );

    DEBUG(('S',3,"out_scan_flo: flo_name == '%s'", flo_name));
    if ( !strcasecmp( (char *) (flo_name + len - 2), "lo" ))
        _each( flo_name, addr, T_ARCMAIL, aso_flavor( flo_name[len - 3] ), _rslow );
    else if ( !strcasecmp( (char *) (flo_name + len - 2), "ut" ))
        _each( flo_name, addr, T_NETMAIL, aso_flavor( flo_name[len - 3] ), _rslow );
    else if ( !strcasecmp( (char *) (flo_name + len - 3), "req" ))
        _each( flo_name, addr, T_REQ, F_REQ, _rslow );
}


/*
 * Returns whether the directory entry is an ASO file.
 */
static int aso_select(OCONST struct dirent *de)
{
    int len;
    char **p;
    FTNADDR_T( a );

    DEBUG(('S',3,"aso_select: '%s'",de->d_name));

    if ( *(de->d_name) == '.' ) {
        return 0;
    }

    if ( strrchr( de->d_name, '.' ) == NULL )
        return 0;

    if ( sscanf( de->d_name, "%hu.%hu.%hu.%hu",
        (UINT16 *) &a.z, (UINT16 *) &a.n, (UINT16 *) &a.f, (UINT16 *) &a.p ) != 4 )
        return 0;

    len = strlen( de->d_name );
    for( p = ext_lo; *p; p++ ) {
        if ( !strncasecmp( &de->d_name[len - 4], *p, 4 )) {
            DEBUG(('S',3,"aso_select: '%s' OK",de->d_name));
            return 1;
        }
    }
    for( p = ext_ut; *p; p++ ) {
        if ( !strncasecmp( &de->d_name[len - 4], *p, 4 )) {
            DEBUG(('S',3,"aso_select: '%s' OK",de->d_name));
            return 1;
        }
    }

    return 0;
}


static int out_aso_rescan(void)
{
    FTNADDR_T( a );
    char flo_name[MAX_PATH];
    struct dirent **namelist;
    int rc = 0, i, n;

    if ( aso_base != NULL && *aso_base != '\0' ) {

        DEBUG(('S',4,"out_aso_rescan: aso_base == '%s'", aso_base));

        n = scandir( aso_base, &namelist, aso_select, NULL );

        for( i = 0; i < n; i++ ) {
            if ( sscanf( namelist[i]->d_name, "%hu.%hu.%hu.%hu",
                (UINT16 *) &a.z, (UINT16 *) &a.n, (UINT16 *) &a.f, (UINT16 *) &a.p ) == 4 ) {

                snprintf( flo_name, MAX_PATH, "%s/%s", aso_base, namelist[i]->d_name );
                DEBUG(('S',4,"out_aso_rescan: flo_name == '%s'", flo_name));
                out_scan_flo( flo_name, &a );
            }
            xfree( namelist[i] );
        }

        xfree( namelist );
        rc = ASO;
    }

    return rc;
}


/*
 * Returns whether the directory entry is a BSO file.
 */
static int bso_select(OCONST struct dirent *de)
{
    int len;
    char **p;

    DEBUG(('S',3,"bso_select: '%s'",de->d_name));

    if ( *(de->d_name) == '.' )
        return 0;

    len = strlen( de->d_name );
    if ( len != 12 )
        return 0;

    if ( strspn( de->d_name, "0123456789abcdefABCDEF" ) != 8 )
        return 0;

    for( p = ext_lo; *p; p++ ) {
        if ( !strncasecmp( &de->d_name[len - 4], *p, 4 )) {
            DEBUG(('S',3,"bso_select: '%s' OK",de->d_name));
            return 1;
        }
    }
    for( p = ext_ut; *p; p++ ) {
        if ( !strncasecmp( &de->d_name[len - 4], *p, 4 )) {
            DEBUG(('S',3,"bso_select: '%s' OK",de->d_name));
            return 1;
        }
    }

    return 0;
}


/*
 * Returns whether the directory entry is a BSO zone.
 */
static int bso_root_select(OCONST struct dirent *de)
{
    int len = strlen( de->d_name );

    DEBUG(('S',3,"bso_root_select: '%s'",de->d_name));

    if ( *(de->d_name) == '.' )
        return 0;

    if ( strncasecmp( de->d_name, p_domain, pd_len ) != 0 )
        return 0;

    if ( len == pd_len ) {
        DEBUG(('S',3,"bso_root_select: '%s' OK",de->d_name));
        return 1;
    }

    if (( len - pd_len ) != 4 )
        return 0;

    if ( strspn( &de->d_name[len - 3], "0123456789abcdefABCDEF" ) != 3 )
        return 0;

    DEBUG(('S',3,"bso_root_select: '%s' OK",de->d_name));
    return 1;
}


static int out_bso_scan_zone(char *zonehold)
{
    int i, j, n, points;
    char base[MAX_PATH], tmp[MAX_PATH], tmp2[MAX_PATH];
    char *p = NULL;
    struct dirent **namelist, **namelist_pnt;
    FTNADDR_T( a );

    DEBUG(('S',3,"out_bso_scan_zone: %s / %s", bso_base, zonehold));

    sprintf( base, "%s/%s", bso_base, zonehold );
    n = scandir( base, &namelist, bso_select, alphasort);

    DEBUG(('S',4,"out_bso_scan_zone: n == %d", n));
    if ( n < 0 ) /* Error */
        return n;

    p = strrchr( zonehold, '.' );
    if ( p != NULL)
        sscanf( p, ".%03hx", (UINT16 *) &a.z );
    else
        a.z = bso_defzone;

    for( i = 0; i < n; i++ ) {
     
        sscanf( namelist[i]->d_name, "%04hx%04hx", (UINT16 *) &a.n, (UINT16 *) &a.f );

        a.p = 0;

        sprintf( tmp, "%s/%s", base, namelist[i]->d_name );
        DEBUG(('S',4,"out_bso_scan_zone: i == %d, lo == '%s'", i, tmp));
        DEBUG(('S',4,"out_bso_scan_zone: %d:%d/%d.%d", a.z, a.n, a.f, a.p));
        if ( !strncasecmp( &(namelist[i]->d_name[8]), ".pnt", 4 )) {
            points = scandir( tmp, &namelist_pnt, bso_select, alphasort );
            if ( points == 0 ) {
                DEBUG(('S',4,"out_bso_scan_zone: remove empty '%s'", tmp));
                rmdir( tmp );
            } else {
                for( j = 0; j < points; j++ ) {
                    sscanf( (char *) (namelist_pnt[j]->d_name + 4), "%04hx", (UINT16 *) &a.p );
                    DEBUG(('S',4,"out_bso_scan_zone: %d:%d/%d.%d", a.z, a.n, a.f, a.p));
                    sprintf( tmp2, "%s/%s", tmp, namelist_pnt[j]->d_name);
                    DEBUG(('S',4,"out_bso_scan_zone: tmp2 == '%s'", tmp2));
                    out_scan_flo( tmp2, &a );
                    xfree( namelist_pnt[j] );
                }
            }
            xfree( namelist_pnt );
        } else {
            out_scan_flo( tmp, &a );
        }

        xfree( namelist[i] );
    }

    xfree( namelist );
    return 1;
}


int out_bso_rescan( void )
{
    int i, n, rc = 0;
    struct dirent **namelist;

    if ( bso_base != NULL && *bso_base != '\0' ) {

        DEBUG(('S',4,"out_bso_rescan: bso_base == '%s', p_domain == '%s'", bso_base, p_domain));

        /* Scan the root and find all outbound directory */
        n = scandir( bso_base, &namelist, bso_root_select, alphasort);
        DEBUG(('S',4,"out_bso_rescan: n == %d", n));

        /* For each directory found do recursive rescan */
        for( i = 0; i < n; i++ ) {
            DEBUG(('S',4,"out_bso_rescan: namelist[%d]->d_name == '%s'", i, namelist[i]->d_name));
            out_bso_scan_zone( namelist[i]->d_name );
            xfree( namelist[i] );
        }
        xfree( namelist );
        rc = BSO;
    }
    return rc;
}


int outbound_rescan(void (*each)(char *, ftnaddr_t *, int, int, int), int rslow)
{
    int rc = 0;

    _each = each;
    _rslow = rslow;

    rc |= out_aso_rescan();
    rc |= out_bso_rescan();

    return rc;
}


int aso_flavor(char fl)
{
    switch( tolower( fl )) {
        case 'h': return F_HOLD;
        case 'f':
        case 'n':
        case 'o': return F_NORM;
        case 'd': return F_DIR;
        case 'c': return F_CRSH;
        case 'i': return F_IMM;
        case 'r': return F_REQ;
    }
    return F_ERR;
}


static int aso_lock(int type,ftnaddr_t *adr,int l)
{
    mkdirs( get_outbound_busy_name( type, adr, 'b' ));
    if ( islocked( tmp ))
        return 0;
    if ( l == LCK_s )
        return lockpid( tmp );
    if ( l == LCK_t )
        return getpid();
    if ( islocked( get_outbound_busy_name( type, adr, 'c' )))
        return 0;
    return lockpid( tmp );
}


/*
 * Returns whether the address is busy. 0 == free, 1 == busy.
 */
int outbound_addr_busy(ftnaddr_t *addr)
{
    if ( islocked( get_outbound_busy_name( ASO, addr, 'b' )))
        return 1;
    if ( islocked( get_outbound_busy_name( ASO, addr, 'c' )))
        return 1;
    if ( islocked( get_outbound_busy_name( BSO, addr, 'b' )))
        return 1;
    if ( islocked( get_outbound_busy_name( BSO, addr, 'c' )))
        return 1;
    return 0;
}


int outbound_locknode(ftnaddr_t *addr, int l)
{
    int rc=0;
    if ( aso_base && aso_lock( ASO, addr, l ))
        rc |= ASO;
    if ( bso_base && aso_lock( BSO, addr, l ))
        rc |= BSO;
    return rc;
}

int outbound_unlocknode(ftnaddr_t *addr, int l)
{
    if ( l == LCK_t )
        return 1;
    if ( aso_base ) {
        lunlink( get_outbound_busy_name( ASO, addr, 'b' ));
        lunlink( get_outbound_busy_name( ASO, addr, 'c' ));
    }
    if ( bso_base ) {
        lunlink( get_outbound_busy_name( BSO, addr, 'b' ));
        lunlink( get_outbound_busy_name( BSO, addr, 'c' ));
        if ( addr->z != bso_defzone )
            rmdirs( tmp );
        else if ( addr->p )
            rmdir( tmp );
    }
    return 1;
}

int aso_attach(ftnaddr_t *addr, int flv, slist_t *files)
{
    slist_t *fl;
    FILE *f=NULL;
    if(bso_base)f=mdfopen(aso_flon(BSO,addr,flv),"at");
    if(!f&&aso_base)f=mdfopen(aso_flon(ASO,addr,flv),"at");
    if(f) {
        for(fl=files;fl;fl=fl->next)fprintf(f,"%s\n",fl->str);
        fclose(f);
        return 1;
    }
    return 0;
}

int aso_request(ftnaddr_t *addr, slist_t *files)
{
    slist_t *fl;
    FILE *f=NULL;
    if(bso_base)f=mdfopen(aso_reqn(BSO,addr),"at");
    if(!f&&aso_base)f=mdfopen(aso_reqn(ASO,addr),"at");
    if(f) {
        for(fl=files;fl;fl=fl->next)fprintf(f,"%s\r\n",fl->str);
        fclose(f);
        return 1;
    }
    return 0;
}

int aso_poll(ftnaddr_t *fa,int flavor)
{
    if(flavor==F_ERR) {
        char *pf=cfgs(CFG_POLLFLAVOR);
        flavor=aso_flavor(*pf);
    }
    return aso_attach(fa,flavor,NULL);
}

int outbound_setstatus(ftnaddr_t *fa,sts_t *st)
{
    FILE *f;
    if((f=mdfopen(aso_stsn(fa),"wt"))) {
        fprintf(f,"%d %d %lu %lu",st->try,st->flags,st->htime,st->utime);
        if(st->bp.name&&st->bp.flags)fprintf(f," %d %d %lu %s",st->bp.flags,st->bp.size,st->bp.time,st->bp.name);
        fclose(f);
        return 1;
    }
    return 0;
}

int outbound_getstatus(ftnaddr_t *fa,sts_t *st)
{
    int rc;
    FILE *f;
    if((f=fopen(aso_stsn(fa),"rt"))) {
        rc=fscanf(f,"%d %d %lu %lu %d %d %lu %s",
            &st->try,&st->flags,(unsigned long*)&st->htime,(unsigned long*)&st->utime,
            &st->bp.flags,(int*)&st->bp.size,(unsigned long*)&st->bp.time,tmp);
        fclose(f);
        if(rc<8)memset(&st->bp,0,sizeof(st->bp));
            else if(*tmp)st->bp.name=xstrdup(tmp);
        if(rc==4||rc==8)return 1;
        write_log("status file %s corrupted",aso_stsn(fa));
        xfree(st->bp.name);
    }
    memset(st,0,sizeof(sts_t));
    return 0;
}

static void floflist(flist_t **fl,char *flon)
{
    FILE *f;
    off_t off;
    char *p,str[MAX_PATH+1],*l,*m,*map=cfgs(CFG_MAPOUT),*fp,*fn;
    slist_t *i;
    struct stat sb;
    int len;
    DEBUG(('S',2,"Add LO '%s'",flon));
    if(!stat(flon, &sb))if((f=fopen(flon,"r+b"))) {
        off=ftell(f);
        while(fgets(str,MAX_PATH,f)) {
            p=strrchr(str,'\r');
            if(!p)p=strrchr(str,'\n');
            if(p)*p=0;
            if(!*str)continue;
            p=str;
            switch(*p) {
                case '~': break;
                case '^': /* kill */
                case '#': /* trunc */
                p++;
                default:
                for(i=cfgsl(CFG_MAPPATH);i;i=i->next) {
                    for(l=i->str;*l&&*l!=' ';l++);
                    for(m=l;*m==' ';m++);
                    len=l-i->str;
                    if(!*l||!*m)write_log("bad mapping '%s'",i->str);
                        else if(!strncasecmp(i->str,p,len)) {
                        memmove(p+strlen(m),p+len,strlen(p+len)+1);
                        memcpy(p,m,strlen(m));
                    }
                }
                if(map&&strchr(map,'S'))strtr(p,'\\','/');
                fp=xstrdup(p);l=basename(fp);
                if(map&&strchr(map,'L'))strlwr(l);
                else if(map&&strchr(map,'U'))strupr(l);
                fn=basename(p);
                    mapname(fn,map,MAX_PATH-(fn-str)-1);
                addflist(fl,fp,xstrdup(fn),str[0],off,f,1);
                if(!stat(fp,&sb)) {
                    totalf+=sb.st_size;totaln++;
                }
            }
            off=ftell(f);
        }
        addflist(fl,xstrdup(flon),NULL,'^',-1,f,1);
    }
}

int asoflist(flist_t **fl,ftnaddr_t *fa,int mode)
{
    struct stat sb;
    char str[MAX_PATH];
    int fls[]={F_IMM,F_CRSH,F_DIR,F_NORM,F_HOLD},i;
    for(i=0;i<5;i++) {
        if(bso_base&&!stat(aso_pktn(BSO,fa,fls[i]),&sb)) {
            snprintf(str,MAX_STRING,"%08lx.pkt",sequencer());
            addflist(fl,xstrdup(tmp),xstrdup(str),'^',0,NULL,0);
            totalm+=sb.st_size;totaln++;
        }
        if(aso_base&&!stat(aso_pktn(ASO,fa,fls[i]),&sb)) {
            snprintf(str,MAX_STRING,"%08lx.pkt",sequencer());
            addflist(fl,xstrdup(tmp),xstrdup(str),'^',0,NULL,0);
            totalm+=sb.st_size;totaln++;
        }
    }
    if(bso_base&&!stat(aso_reqn(BSO,fa),&sb)) {
        snprintf(str,MAX_STRING,"%04x%04x.req",fa->n,fa->f);
        addflist(fl,xstrdup(tmp),xstrdup(str),' ',0,NULL,1);
        totalf+=sb.st_size;totaln++;
    }
    if(aso_base&&!stat(aso_reqn(ASO,fa),&sb)) {
        snprintf(str,MAX_STRING,"%04x%04x.req",fa->n,fa->f);
        addflist(fl,xstrdup(tmp),xstrdup(str),' ',0,NULL,1);
        totalf+=sb.st_size;totaln++;
    }
    for(i=0;i<(5-(cfgi(CFG_HOLDOUT)==1&&mode));i++) {
        if(bso_base)floflist(fl,aso_flon(BSO,fa,fls[i]));
        if(aso_base)floflist(fl,aso_flon(ASO,fa,fls[i]));
    }
    return 0;
}
